/*
 * Copyright 2019 NXP
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted (subject to the limitations in the disclaimer
 * below) provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * * Neither the name of NXP nor the names of its contributors may be used to
 * endorse or promote products derived from this software without specific prior
 * written permission.
 *
 * NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY THIS
 * LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * */

#include <errno.h>
#include <string.h>
#include <sys/syspage.h>
#include <fcntl.h>
#include <stdbool.h>
#include <sys/mman.h>
#include <screen/screen.h>
#include <aoi/aoi.h>
#include <ctype.h>
#include "vdec_cmn.h"
#include "vdec_io.h"


    #include <sys/types.h>
    #include <sys/socket.h>
    #include <netdb.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
#define FRAME_BUFFERS    22
#define POST_DELAY       0

#define PORT_MAX_LEN     10
#define DEFAULT_UDP_PORT 1234
#define READ_BLK_SIZE    9200
#define BYTES_PER_PIXEL  4     /* For RGBA8888 4 bytes per pixel */
#define MAX_LOG_CNT      100

#define dbg_level        LVL_WARN

#define LVL_DBG          10
#define LVL_INFO         3
#define LVL_EVENT        2
#define LVL_WARN         1
#define LVL_ERR          0

#define IMX_VPU_SLOG_OPCODE            _SLOG_SETCODE(_SLOGC_NPAPM,0)
#define IMX_VPU_SLOG_ERROR             _SLOG_ERROR
#define IMX_VPU_SLOG_DEBUG             _SLOG_DEBUG1

#define vpu_dbg(level, fmt, arg...) \
do { \
if (LVL_ERR == (level)) \
slogf(IMX_VPU_SLOG_OPCODE,IMX_VPU_SLOG_ERROR,  fmt, ## arg); \
else if (dbg_level >= (level)) \
slogf(IMX_VPU_SLOG_OPCODE,IMX_VPU_SLOG_DEBUG,  fmt, ## arg); \
} while (0)

int                     quartal = 0;
static vdec_mem_desc_t  sb_mem_desc;
int                     playback_done = 0;

void Play_file(void);


typedef struct screen_buff_info {
    int out_width;
    int out_height;
    int in_width;
    int in_height;
    int frame_res_x;
    int frame_res_y;
} screen_buff_info_t;

typedef struct stats {
    bool                        show_fps;
    uint64_t                    last, now;
    uint64_t                    frames;
    uint64_t                    cps;
    uint64_t                    cond;
} stats_t;

typedef struct screen_ctx {
    screen_window_t screen_bg_win;
    screen_context_t screen_context;
    screen_buffer_t screen_buf[FRAME_BUFFERS];
    off64_t screen_buff_phy[FRAME_BUFFERS];
    size_t screen_buff_size[FRAME_BUFFERS];
    screen_buff_info_t screen_buff_info;
    stats_t stats;
    int     rect[4];
    int dumpcnt;
} screen_ctx_t;

/**
 * Get extension from absolute path.
 *
 * @param filename Absolute path.
 *
 * @return Extension.
 */
const char *get_filename_ext(const char *filename)
{
    const char *dot = strrchr(filename, '.');
    if ((!dot) || (dot == filename)) {
        return "";
    }
    return dot + 1;
}

/**
 * Get video codec from extension
 *
 * @param filename Extension of filename
 *
 * @return Video codec
 */
int32_t get_format_from_extension(const char *filename)
{
    int i = 0;
    char format_str[5] =  {0, 0, 0, 0, 0 };
    const char *extension = get_filename_ext(filename);
    if (strlen(extension) < 3) {
        vpu_dbg(LVL_ERR, "Wrong format len... ");
        return 0;
    }

    for (i = 0; i < strlen(extension); i++) {
        format_str[i] = toupper(extension[i]);
    }

    /* Convert spk suffix to fourcc format */
    if (strcmp(format_str, "SPK") == 0) {
        return (uint32_t)AOIFOURCC("FLV1");
    /* Convert vc1 suffix to fourcc format */
    } else if(strcmp(format_str, "VC1") == 0) {
        return (uint32_t)AOIFOURCC("WVC1");
    }

    return (uint32_t)AOIFOURCC(format_str);
}

/**
 * Init screen.
 *
 * @param scr_ctx Screen context structure.
 *
 * @return 0 if success, -1 if error.
 */
int screen_init(screen_ctx_t *scr_ctx)
{

    scr_ctx->stats.frames = 0;
    scr_ctx->stats.cps = SYSPAGE_ENTRY(qtime)->cycles_per_sec;

    int i = 0;
    int buff_size[2];
    int vis = 0, val = 0;
    int count = 0;
    int dims[2] = { 0, 0 };
    int pos[2] = { 0, 0 };
    int win_swap_interval = 1;
    screen_display_t *screen_disps, screen_disp;
    screen_window_t *screen_bg_win = &scr_ctx->screen_bg_win;
    screen_context_t *screen_context = &scr_ctx->screen_context;

    if (screen_context == NULL) {
        vpu_dbg(LVL_ERR, "Screen context pointer is NULL");
        return -1;
    }

    if (screen_create_context(screen_context, SCREEN_APPLICATION_CONTEXT) < 0) {
        vpu_dbg(LVL_ERR, "Unable to create screen context");
        return -1;
    }
    if (screen_get_context_property_iv(*screen_context, SCREEN_PROPERTY_DISPLAY_COUNT, &count) < 0) {
        vpu_dbg(LVL_ERR, "Unable to get context property");
        goto err0;
    }

    screen_disps = calloc(count, sizeof(screen_display_t));
    if (screen_disps == NULL) {
        vpu_dbg(LVL_ERR, "Display handler is NULL");
        goto err0;
    }

    if (screen_get_context_property_pv(*screen_context, SCREEN_PROPERTY_DISPLAYS, (void **)screen_disps) < 0) {
        vpu_dbg(LVL_ERR, "Unable to get context property");
        goto err1;
    }

    screen_disp = screen_disps[0];

    if (screen_get_display_property_iv(screen_disp, SCREEN_PROPERTY_SIZE, dims) < 0) {
        vpu_dbg(LVL_ERR, "Unable to get display property");
        goto err1;
    }

    scr_ctx->screen_buff_info.out_width = quartal ? dims[0] / 2 : dims[0];
    scr_ctx->screen_buff_info.out_height = quartal ? dims[1] / 2 : dims[1];
    vpu_dbg(LVL_ERR, "SCREEN WXH = %dx%d", scr_ctx->screen_buff_info.out_width, scr_ctx->screen_buff_info.out_height);


    if (screen_create_window(screen_bg_win, *screen_context) < 0) {
        vpu_dbg(LVL_ERR, "Unable to create window");
    }

    if (quartal != 0) {
        pos[0] = (((quartal == 1) || (quartal == 3)) ? 0 : dims[0] / 2);
        pos[1] = (((quartal == 1) || (quartal == 2)) ? 0 : dims[1] / 2);
        dims[0] /= 2;
        dims[1] /= 2;
    }

    if (quartal) {
        if (screen_set_window_property_iv(*screen_bg_win, SCREEN_PROPERTY_POSITION, pos) < 0) {
            vpu_dbg(LVL_ERR, "Unable to get display property");
            goto err2;
        }
    }

    if (quartal) {
        if (screen_set_window_property_iv(*screen_bg_win, SCREEN_PROPERTY_SIZE, dims) < 0) {
            vpu_dbg(LVL_ERR, "Unable to get display property");
            goto err2;
        }
    }

    vis = 0;
    if (screen_set_window_property_iv(*screen_bg_win, SCREEN_PROPERTY_VISIBLE, &vis) < 0) {
        vpu_dbg(LVL_ERR, "Unable to set window property - visible");
        goto err2;
    }

    if (screen_set_window_property_iv(*screen_bg_win, SCREEN_PROPERTY_SWAP_INTERVAL, &win_swap_interval) < 0) {
        vpu_dbg(LVL_ERR, "Unable to set window property - usage");
        goto err2;
    }

    val = SCREEN_USAGE_READ | SCREEN_USAGE_WRITE | SCREEN_USAGE_NATIVE;
    if (screen_set_window_property_iv(*screen_bg_win, SCREEN_PROPERTY_USAGE, &val) < 0) {
        vpu_dbg(LVL_ERR, "Unable to set window property - usage");
        goto err2;
    }

    val = SCREEN_FORMAT_RGBA8888;
    if (screen_set_window_property_iv(*screen_bg_win, SCREEN_PROPERTY_FORMAT, &val) < 0) {
        vpu_dbg(LVL_ERR, "Unable to set window property - format");
        goto err2;
    }

    val = SCREEN_TRANSPARENCY_NONE;
    if (screen_set_window_property_iv(*screen_bg_win, SCREEN_PROPERTY_TRANSPARENCY, &val)) {
        vpu_dbg(LVL_ERR, "Unable to set window property - transparency");
        goto err2;
    }

    if (screen_set_window_property_pv(*screen_bg_win, SCREEN_PROPERTY_DISPLAY, (void **)&screen_disp) < 0) {
        vpu_dbg(LVL_ERR, "Unable to set window property - display");
        goto err2;
    }

    if (screen_create_window_buffers(*screen_bg_win, FRAME_BUFFERS) < 0) {
        vpu_dbg(LVL_ERR, "Unable to create buffers");
        goto err2;
    }

    if (screen_get_window_property_pv(*screen_bg_win, SCREEN_PROPERTY_RENDER_BUFFERS, (void **)&scr_ctx->screen_buf) < 0) {
        vpu_dbg(LVL_ERR, "Unable to set buffer - render buffers: %s", strerror(errno));
        goto err2;
    }


    for (i = 0; i < FRAME_BUFFERS; i++) {

        if (screen_get_buffer_property_llv(scr_ctx->screen_buf[i], SCREEN_PROPERTY_PHYSICAL_ADDRESS,
                                           (long long *)&scr_ctx->screen_buff_phy[i]) < 0) {
            vpu_dbg(LVL_ERR, "Unable to get buffer property - paddr: %s", strerror(errno));
            goto err2;
        }

        if (screen_get_buffer_property_iv(scr_ctx->screen_buf[i], SCREEN_PROPERTY_BUFFER_SIZE, (int *)&buff_size) < 0) {
            vpu_dbg(LVL_ERR, "Unable to get buffer property - paddr: %s", strerror(errno));
            goto err2;
        } else {
            scr_ctx->screen_buff_size[i] = ((size_t)buff_size[0] * (size_t)buff_size[1]);
        }
    }
    vis = 1;
    screen_set_window_property_iv(*screen_bg_win, SCREEN_PROPERTY_VISIBLE, &vis);

    free(screen_disps);

    return 0;

err2:
    screen_destroy_window(*screen_bg_win);
err1:
    free(screen_disps);
err0:
    screen_destroy_context(*screen_context);
    return -1;
}


/**
 * Deinit screen.
 *
 * @param scr_ctx Screen context structure.
 *
 * @return 0 if success, -1 if error.
 */
void screen_deinit(screen_ctx_t *scr_ctx)
{
    screen_destroy_window(scr_ctx->screen_bg_win);
    screen_destroy_context(scr_ctx->screen_context);
}


/**
 * Exit handler.
 *
 * @param dummy.
 */
void exit_handler(int dummy)
{
    vpu_dbg(LVL_ERR, "Player preparing to exit...");
}


/**
 * Append stream data to VPU streambuffer.
 *
 * @param malonefd VPU Handler.
 * @param dec Pointer to exchange structure for driver.
 * @param status Pointer to status structure.
 * @param stream_buffer_virt Stream buffer virtual address.
 * @param buffer Input buffer.
 * @param size Size of stream buffer.
 *
 * @return 0 if success; -1 if an error
 */
int append_buffer(int malonefd, vdec_decode_t *dec, vdec_status_t *status, void *stream_buffer_virt, char *buffer,
                  int size)
{
    unsigned    offset;
    unsigned    head_length;

    offset = status->wptr - status->start;
    head_length = min(status->end - status->wptr, size);

    memcpy(stream_buffer_virt + offset, buffer, head_length);

    if (head_length < size) {
        unsigned tail_length = size - head_length;
        memcpy(stream_buffer_virt, buffer + head_length, tail_length);
    }

    offset = status->wptr - status->start + size;
    dec->wptr = status->start + offset % sb_mem_desc.size;

    return 0;
}

/**
 * Append EOS to VPU streambuffer.
 *
 * @param malonefd VPU Handler.
 * @param dec Pointer to exchange structure for driver.
 * @param init Pointer to init structure.
 * @param status Pointer to status structure.
 * @param sb_virt_uaddr Stream buffer virtual address.
 *
 * @return 0 if success; -1 if an error
 */
static void add_eos(int malonefd, vdec_decode_t *dec, vdec_init_t *init, vdec_status_t *status, void *sb_virt_uaddr)
{
    uint8_t  buffer[READ_BLK_SIZE];
    uint32_t *plbuffer = (uint32_t *)buffer;
    int i = 0;

    if (init->fourcc == (int32_t)AOIFOURCC("H264")) {
        plbuffer[0] = 0x0B010000;
        plbuffer[1] = 0x00000000;
    } else if (init->fourcc == (int32_t)AOIFOURCC("H265") || init->fourcc == (int32_t)AOIFOURCC("HEVC")) {
        plbuffer[0] = 0x4A010000;
        plbuffer[1] = 0x00000020;
    } else if (init->fourcc == (int32_t)AOIFOURCC("MP1V") || init->fourcc == (int32_t)AOIFOURCC("MP2V")) {
        plbuffer[0] = 0xCC010000;
        plbuffer[1] = 0x00000000;
    } else if (init->fourcc == (int32_t)AOIFOURCC("WMV3") || init->fourcc == (int32_t)AOIFOURCC("WVC1")) {
           plbuffer[0] = 0x0A010000;
           plbuffer[1] = 0x00000000;
    } else if (init->fourcc == (int32_t)AOIFOURCC("H263") || init->fourcc == (int32_t)AOIFOURCC("S263")) {
        plbuffer[0] = 0xB1010000;
        plbuffer[1] = 0x00000000;
    } else if (init->fourcc == (int32_t)AOIFOURCC("FLV1")) {
        plbuffer[0] = 0x34010000;
        plbuffer[1] = 0x00000000;
    }

    for (i = 2; i< READ_BLK_SIZE >> 2; i++) {
        plbuffer[i] = 0;
    }

    append_buffer(malonefd, dec, status, sb_virt_uaddr, (void *)buffer, READ_BLK_SIZE);

    vpu_dbg(LVL_WARN, "eos added");

}

/**
 * Dump frame buffer.
 *
 * @param frame Frame index
 * @param ipdata Pointer to framebuffer.
 * @param isize Size of framebuffer to write.
 *
 */
static char fn[] = "/tmp/dump/";
void dump_data(int frame, uint8_t *ipdata, int32_t isize)
{
    FILE *fpindata = NULL;
    char file[1024];
    sprintf(file, "%s/frame%d.raw", fn, frame);
    fpindata = fopen(file, "w");
    if (fpindata == NULL) {
        vpu_dbg(LVL_WARN, "Couldn't %s for dump", file);
        return;
    } else {
        fwrite(ipdata, isize, 1, fpindata);
        vpu_dbg(LVL_WARN, "Dumped frame %d, size %d", frame, isize);
        fclose(fpindata);
    }
}

/**
 * Exit handler.
 *
 * @param s dummy.
 */
void sig_handler(int s)
{
    vpu_dbg(LVL_WARN, "got signal");
    printf("Got signal+++++++\n");
    playback_done = 1;
}


/**
 *
 * @param malonefd VPU Handler.
 * @param screen_ctx Screen context structure.
 * @param oindex Index of output frame buffer.
 * @param clr_idx Index of framebuffer to be cleared.
 *
 * @return 0 if success; -1 if an error
 */
int post_window(int malonefd, screen_ctx_t *screen_ctx, int oindex, int *clr_idx)
{

    usleep(20000);
    if (screen_post_window(screen_ctx->screen_bg_win, screen_ctx->screen_buf[oindex], 1, screen_ctx->rect,
                           SCREEN_WAIT_IDLE)) {
        vpu_dbg(LVL_ERR, "Unable to post window(show_image) with error %s)", strerror(errno));

    } else {
        /* FPS measurement */
        screen_ctx->stats.now = ClockCycles();
        screen_ctx->stats.frames++;
        screen_ctx->stats.cond = screen_ctx->stats.last + screen_ctx->stats.cps;

        if (screen_ctx->stats.now >= screen_ctx->stats.cond) {
            if (screen_ctx->stats.show_fps) {
                printf("[%d]\t FPS=%lu", 0 /* [playerinstance] */, screen_ctx->stats.frames);
            }
            screen_ctx->stats.frames = 0;
            screen_ctx->stats.last = screen_ctx->stats.now;
        }

        vpu_dbg(LVL_DBG, "posted..... frame ( %d) phys 0x%8ld ", oindex, screen_ctx->screen_buff_phy[oindex]);

#if 0 /* DUMP_DATA */
    uint8_t *ptr = NULL;
    /* Get virtual addr (ptr) of screen buffer */
    if (screen_get_buffer_property_pv(screen_ctx->screen_buf[oindex], SCREEN_PROPERTY_POINTER, (void *)&ptr) != EOK) {
        vpu_dbg(LVL_ERR, "screen_get_buffer_property_iv failed for SCREEN_PROPERTY_STRIDE");
    } else {
        /* Dumped data are in BGR32 format, 4bytes per pixel */
        dump_data(screen_ctx->dumpcnt, ptr, screen_ctx->screen_buff_size[oindex] * BYTES_PER_PIXEL);
        screen_ctx->dumpcnt++;
    }
#endif

        if (*clr_idx != -1) {
            /* Screen post window is not really blocking call.
             * Release previous buffer after posting current buffer
             */
            decoder_clear_output(malonefd, *clr_idx);
        }
        *clr_idx = oindex;
    }

    return 0;
}

/**
 * Main rutine.
 *
 * @param argc Number of entered arguments.
 * @param argv Entered arguments.
 *
 * @return 0 if success, otherwise error.
 */
int main(int argc, char *argv[])
{
    Play_file();
    //Play_file();
       return 0;
}

void Play_file()
{
     int                         opt_index = 0;
    int                         i = 0;

    /* Stream and stream buffer */
    char                        clip_path[128];
    int                         stream_fd;
    void                        *sb_virt_uaddr;
    int32_t                     sb_fd;
    //char                        buffer[READ_BLK_SIZE];
    char *buffer;
    int                         len;
    int                         space, delta;

    /* Screen */
    screen_ctx_t                screen_ctx;

    /* Frame buffers */
    frame_buffer_t              fb[FRAME_BUFFERS];
    frame_buffers_t             fbufs;
    int                         clr_idx = -1;

    /* Decoder */
    vdec_init_t                 init;
    vdec_decode_t               dec;
    vdec_status_t               status;
    int                         malonefd = 0;
    bool                        format_set_flag = 0;

    /* UDP Socket */
    int                         sockfd = 0;
    int                         connfd = 0;
    struct sockaddr_in          servaddr, cliaddr;
    int                         port = DEFAULT_UDP_PORT;
    int                         network_stream = 0;
    socklen_t                   addrlen = sizeof(cliaddr);



            int iTempLen = 0;
            FILE *fp;
            fp = fopen("/tmp/frame_length.txt","rb");


    memset(&dec, 0, sizeof(dec));

    memset(&servaddr, 0, sizeof(servaddr));
    memset(&cliaddr, 0, sizeof(cliaddr));

    memset(clip_path, 0, 128);
    screen_ctx.stats.show_fps = true;
    memset(&screen_ctx, 0, sizeof(screen_ctx_t));


    signal(SIGINT, sig_handler);
    signal(SIGKILL, sig_handler);
    signal(SIGTERM, sig_handler);

    /* Check input variables */
    strcpy(clip_path, "/tmp/video_data.h264");
    init.fourcc = get_format_from_extension(clip_path);

    {
        /* LOAD FILE */
        printf( "Loading file %s...\n", clip_path);
        stream_fd = open(clip_path, O_RDONLY);
        if (stream_fd == -1) {
            printf("File %s not found!\n", clip_path);
            return EXIT_FAILURE;
        }
    }

    /* Screen init */
    if (screen_init(&screen_ctx) != 0) {
        vpu_dbg(LVL_ERR, "%s: Unable to init screen. Screen is not running or crashed.\n ", __FUNCTION__);
        goto exit0;
    }

    screen_ctx.rect[0] = 0;
    screen_ctx.rect[1] = 0;
    screen_ctx.rect[2] = screen_ctx.screen_buff_info.out_width;
    screen_ctx.rect[3] = screen_ctx.screen_buff_info.out_height;


    /* Malone open */
    malonefd = decoder_open();
    if (malonefd == -1) {
        vpu_dbg(LVL_ERR, "decoder_open failed");
        goto exit1;
    }

    /* Disable reorder for low latency mode*/
     init.dis_reorder = 1; 

    /* Malone init */
    if (decoder_init(malonefd, &init, &sb_mem_desc) != EOK) {
        vpu_dbg(LVL_ERR, "decoder_init failed");
        goto exit2;
    }

    /* Stream buffer init */
    sb_fd = shm_open(sb_mem_desc.memobj_path, O_RDWR, 0777);
    if (sb_fd == -1) {
        vpu_dbg(LVL_ERR, "Stream buffer shm_open failed");
        goto exit3;
    }

    sb_virt_uaddr  = mmap64(0, sb_mem_desc.size, PROT_WRITE | PROT_NOCACHE, MAP_NOINIT | MAP_SHARED, sb_fd, 0);

    if (sb_virt_uaddr == MAP_FAILED) {
        vpu_dbg(LVL_ERR, "%s Unable to open stream buffer shared memory: %s", __func__, strerror(errno));
        goto exit4;
    }

    vpu_dbg(LVL_INFO, "virt %p size %d path=%s", (void *)sb_virt_uaddr, sb_mem_desc.size, sb_mem_desc.memobj_path);

    /* Must be called to reinitialize codec instance */
    if (decoder_start(malonefd) != EOK) {
        vpu_dbg(LVL_ERR, "%s decoder_start failed", __func__);
        goto exit5;
    }

    if (decoder_status(malonefd, &status) != EOK) {
        vpu_dbg(LVL_ERR, "%s decoder_status failed", __func__);
        goto exit6;
    }

    dec.wptr = status.wptr;
   // dec.timeout = 0;
    dec.timeout = 30;

    /* Register frame buffers */
    for (i = 0; i < FRAME_BUFFERS; i++) {

        int ostride;
        if (screen_get_buffer_property_iv(screen_ctx.screen_buf[i], SCREEN_PROPERTY_STRIDE, &ostride) != EOK) {
            vpu_dbg(LVL_ERR, "screen_get_buffer_property_iv failed for SCREEN_PROPERTY_STRIDE");
            goto exit6;
        }

        fb[i].phy_addr  = screen_ctx.screen_buff_phy[i];
        fb[i].index = i;
        fb[i].size = screen_ctx.screen_buff_size[i];
        fb[i].ostride = ostride;
        fb[i].owidth  = screen_ctx.screen_buff_info.out_width;
        fb[i].oheight = screen_ctx.screen_buff_info.out_height;
    }

    fbufs.fb = fb;
    fbufs.count = FRAME_BUFFERS;

    if (decoder_register_frame_buffer(malonefd, &fbufs) != EOK) {
        vpu_dbg(LVL_ERR,  "decoder_register_frame_buffer Failed");
        goto exit6;
    }

    /*
     * DECODE
     */
printf("stream_fd is = %d sockfd =%d playback_done =%d\n",stream_fd,sockfd,playback_done);
    while ((stream_fd >= 0 || sockfd > 0) && (!playback_done)) {

        if ((sockfd > 0) && (connfd == 0)) {

            if ((listen(sockfd, 5)) != 0) {
                vpu_dbg(LVL_ERR,  "Socket listen failed.");
                goto exit6;
            } else {
                vpu_dbg(LVL_WARN, "TCP listening... \n");
            }

            connfd = accept(sockfd, (struct sockaddr*)&cliaddr, &addrlen);
            if (connfd < 0) {
                vpu_dbg(LVL_ERR, "Server accept failed ...\n");
                goto exit6;
            } else {
                vpu_dbg(LVL_WARN, "Server accept the client ...\n");
            }
        }

        if (decoder_status(malonefd, &status) != EOK) {
            vpu_dbg(LVL_ERR, "decoder_status error");
            break;
        }
        if (status.stopped) {
            vpu_dbg(LVL_ERR, "stopped");
            playback_done = true;
            printf("PLAYBACK DONE 1111 +++++++++++++++++++++\n");
            break;
        }

        delta = status.rptr - status.wptr;
        space = delta > 0 ? delta : sb_mem_desc.size + delta;
#if 0
        if (space > sizeof(buffer)) {
            len = 0;
            /* Read more data if space is available */
            if (stream_fd >= 0 ) {
                len = read(stream_fd, buffer, sizeof(buffer));

            } else if (connfd > 0) {
                len = read(connfd, buffer, sizeof(buffer));
            }
#endif



            fscanf(fp, "%d", &iTempLen);
            printf("frame len readed is %d\n",iTempLen);
           // fclose(fp);
            buffer = (char *)calloc(1, iTempLen+4096);
            if(buffer == NULL)
            {
                printf("calloc failure\n");
            }
            if (space > iTempLen) {
            len = 0;
            /* Read more data if space is available */
            if (stream_fd > 0 ) {
                len = read(stream_fd, buffer, iTempLen);
                printf("fread len = %d\n",len);
                char cPrefix[4];
                cPrefix[0] = 0;
                cPrefix[1] = 0;
                cPrefix[2] = 1;
                cPrefix[3] = 21;
                memcpy(&buffer[iTempLen],cPrefix,4);
                memset(&buffer[iTempLen + 4], 0, 4092);
#if 0
                FILE *fppp;
                fppp = fopen("Lowlatency.h264","ab");
                fwrite(buffer,(iTempLen+4096), 1 ,fppp);
                fclose(fppp);
#endif 

            } else if (connfd > 0) {
                len = read(connfd, buffer, iTempLen);
            }



           // printf("data readed is = %d\n",len);
            switch (len) {
                case -1:
                    vpu_dbg(LVL_ERR,"Error reading input file");
                    if (sockfd > 0) {
                        close(sockfd);
                        sockfd = -1;
                        dec.eos = 1;
                        vpu_dbg(LVL_INFO, "input eos");
                        add_eos(malonefd, &dec, &init, &status, sb_virt_uaddr);
                    }
                    break;
                case 0:
                    if (network_stream) {
                        /* Example of seek operation, but here
                         * it is used for loosing client connection */
                        decoder_stop(malonefd, 1);
                        for (i = 0; i < FRAME_BUFFERS; i++) {
                            decoder_clear_output(malonefd, i);
                        }
                        dec.wptr = status.start;
                        decoder_start(malonefd);
                        connfd = 0;
                    } else {
                        close(stream_fd);
                        stream_fd = -1;
                    }
                    dec.eos = 1;
                    vpu_dbg(LVL_INFO, "input eos");
                    add_eos(malonefd, &dec, &init, &status, sb_virt_uaddr);
                    break;
                default:
                    /* Update wptr */
                    append_buffer(malonefd, &dec, &status, sb_virt_uaddr, buffer, len + 4096);
                    break;
            }
        }

        /* Malone decode call */
        printf("before decode \n");
        dec.frame_in_cnt++;
        decoder_decode(malonefd, &dec);
        //printf("width and height is = %d= %d\n",dec.width,dec.heigth);
        printf("dec.oindex is = %d\n",dec.oindex);
        if (dec.oindex != -1) {
            post_window(malonefd, &screen_ctx, dec.oindex, &clr_idx);
        } else {
            usleep(POST_DELAY);
        }
    }

    while (!playback_done) {
        vdec_status_t status;
        dec.eos = 1;

        /* Malone decode call */
        decoder_decode(malonefd, &dec);
        printf("decoder last frames index is = %d\n",dec.oindex);
        if (dec.oindex == -1) 
        {
              printf("DEcoder Last frames oindex is = %d\n",dec.oindex);
                  
        }
        if (dec.oindex != -1) {
            post_window(malonefd, &screen_ctx, dec.oindex, &clr_idx);
            usleep(20000);
        } else if (decoder_status(malonefd, &status) != EOK) {
            vpu_dbg(LVL_ERR, "decoder_status error");
            break;
        } else if (status.eos) {
            vpu_dbg(LVL_DBG, "EOS");
            break;
        } else if (status.stopped) {
            vpu_dbg(LVL_ERR, "stopped");
            playback_done = true;
            printf("PLAYBACK DONE +++++++++++++++++++++\n");
            break;
        } else {
            usleep(POST_DELAY);
        }
    }

    printf("before decoder_clear_output \n");
exit6:
    if (clr_idx != -1) {
        decoder_clear_output(malonefd, clr_idx);
    }
    decoder_stop(malonefd, 1); /* Must be blocking call */
exit5:
    munmap(sb_virt_uaddr, sb_mem_desc.size);
exit4:
    close(sb_fd);
exit3:
    decoder_deinit(malonefd);
exit2:
    decoder_close(malonefd);
exit1:
    screen_deinit(&screen_ctx);
exit0:
    close(stream_fd);
    close(connfd);
    printf("before return \n");
return;
}

